#include <stream.h>
#include <math.h>

#include "ABC.h"  // now this is OK for noisy and not noisy


abcfilter::abcfilter( filter_params parms)
{
int i;

f_params = parms;   //this doesn't copy, I guess...  so caller controls these vital
			// parms? ugh

activity = initializing;
age_ticks = 0;
age_secs = 0.0;
data_ticks = 0;
data_secs = 0.0;
nodata_ticks = 0;
nodata_secs = 0.0;
timeout_ticks = irint(f_params.timeout_secs/f_params.time_step);

xcurr = vcurr = vave = xave = aave =  acurr= 0.0;

}


void abcfilter::run(data_type signal, double X, double advance)
{

activity = (age_ticks < f_params.init_ticks) ?  initializing : tracking;
		//tracking or predicting actually but we'll find that out later
		
	double adv =  advance;

// cout << form("\n ticknoise: %f", tick_noise);

if ( activity == initializing)
	{

	if (signal == nodata) {activity= timed_out; return;} //screw it for now;
		//policy decision -- don't initialize through dropout

	N = double(age_ticks) + 1.0;  //for running average in itialization

/* there follows a segment to initialize the filter..keeps a runnning ave.
of the velocity, uses latest x value.   This idea is appropriate for const
velocity but may not be for any other velocity */

	 if (age_ticks >= 0)
		 { xlast = xcurr;
		xcurr = X;
		}

	 if (age_ticks >= 1)
		 {
		 vlast = vcurr;
		  vcurr = (xcurr - xlast)/f_params.time_step;
		 // vave = vave + (vcurr - vave)/(N-1.0);  useless, right?
		 }
	if (age_ticks >= 2)
                        {
                        alast = acurr;
                        acurr = (vcurr - vlast)/f_params.time_step;
                        aave = aave + (acurr - aave)/(N-2.0);
                        }


	if (age_ticks < 2) activity = invalid; 

/* now compute the estimates during initialization */

		 x_est = xcurr;
//		 v_est = vcurr - 0.5 * aave*f_params.time_step;
		 v_est = vcurr;
		a_est = aave;
		 x_future = x_est + v_est*(adv) +
			+ 0.5 * a_est * (adv*adv);
		 v_future = v_est + adv*a_est; 
		a_future = a_est;

		}	
		//end initialization phase

	else  //filtering
	{

/* start timer if no data...*/

	if (signal == nodata)
		{
		if (nodata_ticks++ >= timeout_ticks) 
			{activity= timed_out; return;}
		activity = predicting;
		data_ticks = 0;

		x_est = x_est + v_est*f_params.time_step + 0.5*a_est*f_params.time_step*f_params.time_step;
		 //predict based on last known vel, acc..
		v_est = v_est + a_est*f_params.time_step ;  //last known acc...
		}

/* do you believe it?  the alpha-beta equations! */

	if (signal == data)
		{
		activity =  tracking;
		nodata_ticks = 0;
		data_ticks++;
		x_pred = x_est + f_params.time_step*v_est  + 0.5*a_est*f_params.time_step*f_params.time_step;
		v_pred = v_est + f_params.time_step*a_est;
		innov = X - x_pred;
		x_est = x_pred + f_params.alpha*innov;
		v_est = v_pred + (f_params.beta/f_params.time_step)*innov;
		a_est = a_est + (f_params.gamma/(f_params.time_step*f_params.time_step))*innov;

		}

 		x_future = x_est + v_est*adv + 0.5*a_est*adv*adv;
                v_future = v_est + adv*a_est;
		a_future = a_est;

	}

age_ticks++;  //go to next time step
age_secs += f_params.time_step;  // in real time...
data_secs = data_ticks*f_params.time_step;
nodata_secs = nodata_ticks*f_params.time_step;


}  // end run


void abcfilter::dump(void)
{
cout << form("\n\n ABC FILTER PARAMETERS:");
cout << form("\n init_ticks, timeout_secs, timeout_ticks: %d %f %d ",
		f_params.init_ticks, f_params.timeout_secs, timeout_ticks);
cout << form("\n alpha, beta, gamma: %f %f %f",f_params.alpha, f_params.beta,
			f_params.gamma);
cout << form("\n ");

cout << form("\n\n FILTER STATE:");
cout << form("\n age_ticks, data_ticks, no_data_ticks: %d %d %d",
	age_ticks,  data_ticks, nodata_ticks);
cout << form("\n age_secs, data_secs, no_data_secs: %f %f %f",
	age_secs,  data_secs, nodata_secs);
cout << form("\n filter activity: ");
	switch (activity)
	{
	case invalid: cout << " invalid";
	break;
	case initializing: cout << " initializing";
	break;
	case tracking: cout << " tracking";
	break;
	case predicting: cout << " predicting";
	break;
	case timed_out: cout << " timed_out";
	break;
	}

cout << form("\n x_est, v_est, a_est:  %f %f %f",
	x_est,v_est, a_est);
cout << form("\n x_future, v_future, a_future: %f %f %f  ",
	x_future, v_future, a_future);
}

/*here is an overload of run, which takes the current timestamp and does
the same number as the above -- a variable timestep ABC. hope it works...*/
/* here a "tick" is just number of times the sucker has been called...nothing
        to do with time */


void abcfilter::run(data_type signal, double X, double advance, double time_stamp)
{

double this_step = time_stamp - age_secs;

activity = (age_ticks < f_params.init_ticks) ?  initializing : tracking;
		//tracking or predicting actually but we'll find that out later
		
	double adv =  advance;

// cout << form("\n ticknoise: %f", tick_noise);

if ( activity == initializing)
	{

	if (signal == nodata) {activity= timed_out; return;} 
		//policy decision -- don't initialize through dropout

	N = double(age_ticks) + 1.0;  //for running average in itialization
	// this use of N means we trust the info from the initial
        // ticks equally (regardless of their separation in time, say

/* there follows a segment to initialize the filter..keeps a runnning ave.
of the velocity, uses latest x value.   This idea is appropriate for const
velocity but may not be for any other velocity */

	 if (age_ticks >= 0)
		 { xlast = xcurr;
		xcurr = X;
		}

	 if (age_ticks >= 1)
		 {
		 vlast = vcurr;
		  vcurr = (xcurr - xlast)/this_step;
		 // vave = vave + (vcurr - vave)/(N-1.0);  useless, right?
		 }
	if (age_ticks >= 2)
                        {
                        alast = acurr;
                        acurr = (vcurr - vlast)/this_step;
                        aave = aave + (acurr - aave)/(N-2.0);
                        }


	if (age_ticks < 2) activity = invalid; 

/* now compute the estimates during initialization */

		 x_est = xcurr;
//		 v_est = vcurr - 0.5 * aave*this_step;
		 v_est = vcurr;
		a_est = aave;
		 x_future = x_est + v_est*(adv) +
			+ 0.5 * a_est * (adv*adv);
		 v_future = v_est + adv*a_est; 
		a_future = a_est;

		}	
		//end initialization phase

	else  //filtering
	{

/* start timer if no data...*/

	if (signal == nodata)
		{
		if ((nodata_secs += this_step) >= timeout_ticks) 
			{activity= timed_out; return;}
		activity = predicting;
		nodata_ticks++;
		data_ticks = 0;
		data_secs = 0.0;

		x_est = x_est + v_est*this_step + 0.5*a_est*this_step*this_step;
		 //predict based on last known vel, acc..
		v_est = v_est + a_est*this_step ;  //last known acc...
		}

/* do you believe it?  the alpha-beta equations! */

	if (signal == data)
		{
		activity =  tracking;
		nodata_ticks = 0;
		nodata_secs = 0.0;
		data_ticks++;
		data_secs+= this_step;
		x_pred = x_est + this_step*v_est  + 0.5*a_est*this_step*this_step;
		v_pred = v_est + this_step*a_est;
		innov = X - x_pred;
		x_est = x_pred + f_params.alpha*innov;
		v_est = v_pred + (f_params.beta/this_step)*innov;
		a_est = a_est + (f_params.gamma/(this_step*this_step))*innov;

		}

 		x_future = x_est + v_est*adv + 0.5*a_est*adv*adv;
                v_future = v_est + adv*a_est;
		a_future = a_est;

	}

age_ticks++;  //go to next time step
age_secs = time_stamp; 

}  // end run

